
//
// This source file is part of appleseed.
// Visit http://appleseedhq.net/ for additional information and resources.
//
// This software is released under the MIT license.
//
// Copyright (c) 2010-2013 Francois Beaune, Jupiter Jazz Limited
// Copyright (c) 2014 Francois Beaune, The appleseedhq Organization
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef APPLESEED_RENDERER_KERNEL_SHADING_SHADINGPOINT_H
#define APPLESEED_RENDERER_KERNEL_SHADING_SHADINGPOINT_H

// appleseed.renderer headers.
#include "renderer/global/globaltypes.h"
#include "renderer/kernel/intersection/intersectionsettings.h"
#include "renderer/kernel/shading/shadingray.h"
#include "renderer/kernel/tessellation/statictessellation.h"
#include "renderer/modeling/material/inormalmodifier.h"
#include "renderer/modeling/material/material.h"
#include "renderer/modeling/object/curveobject.h"
#include "renderer/modeling/object/regionkit.h"
#include "renderer/modeling/object/triangle.h"
#include "renderer/modeling/scene/assembly.h"
#include "renderer/modeling/scene/assemblyinstance.h"
#include "renderer/modeling/scene/containers.h"
#include "renderer/modeling/scene/objectinstance.h"

// appleseed.foundation headers.
#include "foundation/math/basis.h"
#include "foundation/math/beziercurve.h"
#include "foundation/math/transform.h"
#include "foundation/math/vector.h"
#include "foundation/platform/compiler.h"
#include "foundation/platform/types.h"

// OSL headers.
#ifdef WITH_OSL
#include "foundation/platform/oslheaderguards.h"
BEGIN_OSL_INCLUDES
#include "OSL/shaderglobals.h"
END_OSL_INCLUDES
#endif

// Standard headers.
#include <cassert>
#include <cstddef>

// Forward declarations.
namespace renderer  { class Object; }
#ifdef WITH_OSL
namespace renderer  { class OSLShaderGroupExec; }
#endif
namespace renderer  { class Scene; }
namespace renderer  { class TextureCache; }

namespace renderer
{

//
// A shading point.
//

class ShadingPoint
{
  public:
    // The type of the primitive at the shading point.
    enum PrimitiveType
    {
        PrimitiveNone,
        PrimitiveTriangle,
        PrimitiveCurve
    };

    // Constructor, calls clear().
    ShadingPoint();

    // Copy constructor.
    explicit ShadingPoint(const ShadingPoint& rhs);

    // Reset the shading point to its initial state (no intersection).
    void clear();

    // Replace the ray stored in the shading point.
    void set_ray(const ShadingRay& ray);

    // Return the scene that was tested for intersection.
    const Scene& get_scene() const;

    // Return the world space ray that was cast through the scene.
    const ShadingRay& get_ray() const;

    // Return the time stored in the ray.
    double get_time() const;

    // Return true if an intersection was found, false otherwise.
    bool hit() const;

    // Return the type of the hit primitive.
    PrimitiveType get_primitive_type() const;

    // Return the distance from the ray origin to the intersection point.
    double get_distance() const;

    // Return the barycentric coordinates of the intersection point.
    const foundation::Vector2d& get_bary() const;

    // Return the texture coordinates from a given UV set at the intersection point.
    const foundation::Vector2d& get_uv(const size_t uvset) const;

    // Return the intersection point in world space.
    const foundation::Vector3d& get_point() const;

    // Return the intersection point in world space, with per-object-instance ray bias applied.
    foundation::Vector3d get_biased_point(const foundation::Vector3d& direction) const;

    // Return the intersection point in world space, properly offset to avoid self-intersections.
    const foundation::Vector3d& get_offset_point(const foundation::Vector3d& direction) const;

    // Return the world space partial derivatives of the intersection point wrt. a given UV set.
    const foundation::Vector3d& get_dpdu(const size_t uvset) const;
    const foundation::Vector3d& get_dpdv(const size_t uvset) const;

    // Return the world space geometric normal at the intersection point. The geometric normal
    // always faces the incoming ray, i.e. dot(ray_dir, geometric_normal) is always positive or null.
    const foundation::Vector3d& get_geometric_normal() const;

    // Return the world space (possibly modified) shading normal at the intersection point.
    // The shading normal is always in the same hemisphere as the geometric normal, but it is
    // not always facing the incoming ray, i.e. dot(ray_dir, shading_normal) may be negative.
    const foundation::Vector3d& get_shading_normal() const;

    // Return the original world space shading normal at the intersection point.
    const foundation::Vector3d& get_original_shading_normal() const;

    // Return a world space orthonormal basis around the (possibly modified) shading normal.
    const foundation::Basis3d& get_shading_basis() const;

    // Return the side of the surface that was hit.
    ObjectInstance::Side get_side() const;

    // Return the i'th world space vertex of the hit triangle.
    const foundation::Vector3d& get_vertex(const size_t i) const;

    // Return the world space normal at the i'th vertex of the hit triangle.
    const foundation::Vector3d& get_vertex_normal(const size_t i) const;

    // Return the material at the intersection point, or 0 if there is none.
    const Material* get_material() const;

    // Return the assembly instance that was hit.
    const AssemblyInstance& get_assembly_instance() const;

    // Return the assembly that was hit.
    const Assembly& get_assembly() const;

    // Return the object instance that was hit.
    const ObjectInstance& get_object_instance() const;

    // Return the object that was hit.
    const Object& get_object() const;

    // Return the index, within the assembly, of the object instance that was hit.
    size_t get_object_instance_index() const;

    // Return the index, within the object, of the region containing the hit triangle.
    size_t get_region_index() const;

    // Return the index of the hit primitive.
    size_t get_primitive_index() const;

    // Return the index of the primitive attribute.
    size_t get_primitive_attribute_index() const;

#ifdef WITH_OSL
    struct OSLObjectTransformInfo
    {
        bool is_animated() const;

        OSL::Matrix44 get_transform() const;
        OSL::Matrix44 get_transform(float t) const;

        OSL::Matrix44 get_inverse_transform() const;
        OSL::Matrix44 get_inverse_transform(float t) const;

        const TransformSequence*        m_assembly_instance_transform;
        const foundation::Transformd*   m_object_instance_transform;
    };

    struct OSLTraceData
    {
        bool        m_traced;
        bool        m_hit;
        float       m_hit_distance;
        OSL::Vec3   m_P;
        OSL::Vec3   m_N;
        OSL::Vec3   m_Ng;
        float       m_u;
        float       m_v;
    };

    OSL::ShaderGlobals& get_osl_shader_globals() const;
#endif

  private:
    friend class AssemblyLeafProbeVisitor;
    friend class AssemblyLeafVisitor;
    friend class CurveLeafVisitor;
    friend class Intersector;
#ifdef WITH_OSL
    friend class OSLShaderGroupExec;
#endif
    friend class RegionLeafVisitor;
    friend class ShadingPointBuilder;
    friend class TriangleLeafVisitor;

    RegionKitAccessCache*               m_region_kit_cache;
    StaticTriangleTessAccessCache*      m_tess_cache;
    TextureCache*                       m_texture_cache;

    const Scene*                        m_scene;
    mutable ShadingRay                  m_ray;                          // world space ray (m_tmax = distance to intersection)

    // Intersection results.
    PrimitiveType                       m_primitive_type;               // type of the hit primitive
    foundation::Vector2d                m_bary;                         // barycentric coordinates of intersection point
    const AssemblyInstance*             m_assembly_instance;            // hit assembly instance
    foundation::Transformd              m_assembly_instance_transform;  // transform of the hit assembly instance at ray time
    size_t                              m_object_instance_index;        // index of the object instance that was hit
    size_t                              m_region_index;                 // index of the region containing the hit triangle
    size_t                              m_primitive_index;              // index of the hit primitive
    TriangleSupportPlaneType            m_triangle_support_plane;       // support plane of the hit triangle

    // Additional intersection results, computed on demand.
    enum Members
    {
        HasSourceGeometry               = 1 << 0,
        HasUV0                          = 1 << 1,
        HasPoint                        = 1 << 2,
        HasBiasedPoint                  = 1 << 3,
        HasRefinedPoints                = 1 << 4,
        HasPartialDerivatives           = 1 << 5,
        HasGeometricNormal              = 1 << 6,
        HasShadingNormal                = 1 << 7,
        HasOriginalShadingNormal        = 1 << 8,
        HasShadingBasis                 = 1 << 9,
        HasWorldSpaceVertices           = 1 << 10,
        HasWorldSpaceVertexNormals      = 1 << 11,
        HasMaterial                     = 1 << 12

#ifdef WITH_OSL
        , HasOSLShaderGlobals           = 1 << 13
#endif
    };
    mutable foundation::uint32          m_members;                      // which members have already been computed
    mutable const Assembly*             m_assembly;                     // hit assembly
    mutable const ObjectInstance*       m_object_instance;              // hit object instance
    mutable Object*                     m_object;                       // hit object
    mutable foundation::uint32          m_primitive_pa;                 // hit primitive attribute index
    mutable GVector2                    m_v0_uv, m_v1_uv, m_v2_uv;      // texture coordinates from UV set #0 at triangle vertices
    mutable GVector3                    m_v0, m_v1, m_v2;               // object instance space triangle vertices
    mutable GVector3                    m_n0, m_n1, m_n2;               // object instance space triangle vertex normals
    mutable foundation::Vector2d        m_uv;                           // texture coordinates from UV set #0
    mutable foundation::Vector3d        m_point;                        // world space intersection point
    mutable foundation::Vector3d        m_biased_point;                 // world space intersection point with per-object-instance bias applied
    mutable foundation::Vector3d        m_dpdu;                         // world space partial derivative of the intersection point wrt. U
    mutable foundation::Vector3d        m_dpdv;                         // world space partial derivative of the intersection point wrt. V
    mutable foundation::Vector3d        m_geometric_normal;             // world space geometric normal, unit-length
    mutable foundation::Vector3d        m_shading_normal;               // world space (possibly modified) shading normal, unit-length
    mutable foundation::Vector3d        m_original_shading_normal;      // original world space shading normal, unit-length
    mutable ObjectInstance::Side        m_side;                         // side of the surface that was hit
    mutable foundation::Basis3d         m_shading_basis;                // world space orthonormal basis around shading normal
    mutable foundation::Vector3d        m_v0_w, m_v1_w, m_v2_w;         // world space triangle vertices
    mutable foundation::Vector3d        m_n0_w, m_n1_w, m_n2_w;         // world space triangle vertex normals
    mutable const Material*             m_material;                     // material at intersection point

    // Data required to avoid self-intersections.
    mutable foundation::Vector3d        m_asm_geo_normal;               // assembly instance space geometric normal to hit triangle
    mutable foundation::Vector3d        m_front_point;                  // hit point refined to front, in assembly instance space
    mutable foundation::Vector3d        m_back_point;                   // hit point refined to back, in assembly instance space

#ifdef WITH_OSL
    mutable OSLObjectTransformInfo      m_obj_transform_info;
    mutable OSLTraceData                m_osl_trace_data;
    mutable OSL::ShaderGlobals          m_shader_globals;
#endif

    // Cache the source geometry, fetching it if necessary.
    void cache_source_geometry() const;

    // Fetch the source geometry.
    void fetch_source_geometry() const;

    // Additional fetch functions for different primitive types
    void fetch_triangle_source_geometry() const;

    void fetch_curve_source_geometry() const;

    // Refine and offset the intersection point.
    void refine_and_offset() const;

    // Compute the partial derivatives dp/du and dp/dv.
    void compute_partial_derivatives() const;
};


//
// ShadingPoint class implementation.
//

FORCE_INLINE ShadingPoint::ShadingPoint()
{
    clear();
}

inline ShadingPoint::ShadingPoint(const ShadingPoint& rhs)
  : m_region_kit_cache(rhs.m_region_kit_cache)
  , m_tess_cache(rhs.m_tess_cache)
  , m_texture_cache(rhs.m_texture_cache)
  , m_scene(rhs.m_scene)
  , m_ray(rhs.m_ray)
  , m_primitive_type(rhs.m_primitive_type)
  , m_bary(rhs.m_bary)
  , m_assembly_instance(rhs.m_assembly_instance)
  , m_assembly_instance_transform(rhs.m_assembly_instance_transform)
  , m_object_instance_index(rhs.m_object_instance_index)
  , m_region_index(rhs.m_region_index)
  , m_primitive_index(rhs.m_primitive_index)
  , m_triangle_support_plane(rhs.m_triangle_support_plane)
  , m_members(0)
{
}

FORCE_INLINE void ShadingPoint::clear()
{
    m_region_kit_cache = 0;
    m_tess_cache = 0;
    m_texture_cache = 0;
    m_scene = 0;
    m_primitive_type = PrimitiveNone;
    m_members = 0;
}

inline void ShadingPoint::set_ray(const ShadingRay& ray)
{
    m_ray = ray;
}

inline const Scene& ShadingPoint::get_scene() const
{
    assert(m_scene);
    return *m_scene;
}

inline const ShadingRay& ShadingPoint::get_ray() const
{
    return m_ray;
}

inline double ShadingPoint::get_time() const
{
    return m_ray.m_time;
}

inline bool ShadingPoint::hit() const
{
    return m_primitive_type != PrimitiveNone;
}

inline ShadingPoint::PrimitiveType ShadingPoint::get_primitive_type() const
{
    return m_primitive_type;
}

inline double ShadingPoint::get_distance() const
{
    assert(hit());
    return m_ray.m_tmax;
}

inline const foundation::Vector2d& ShadingPoint::get_bary() const
{
    assert(hit());
    return m_bary;
}

inline const foundation::Vector2d& ShadingPoint::get_uv(const size_t uvset) const
{
    assert(hit());
    assert(uvset == 0);     // todo: support multiple UV sets

    if (!(m_members & HasUV0))
    {
        cache_source_geometry();

        if (m_primitive_type == PrimitiveTriangle)
        {
            // Compute the texture coordinates.
            const foundation::Vector2d v0_uv(m_v0_uv);
            const foundation::Vector2d v1_uv(m_v1_uv);
            const foundation::Vector2d v2_uv(m_v2_uv);
            const double w = 1.0 - m_bary[0] - m_bary[1];
            m_uv =
                  v0_uv * w
                + v1_uv * m_bary[0]
                + v2_uv * m_bary[1];
        }
        else
        {
            assert(m_primitive_type == PrimitiveCurve);
            m_uv = m_bary;
        }

        // Texture coordinates from UV set #0 are now available.
        m_members |= HasUV0;
    }

    return m_uv;
}

inline const foundation::Vector3d& ShadingPoint::get_point() const
{
    assert(hit());

    if (!(m_members & HasPoint))
    {
        m_point = m_ray.point_at(m_ray.m_tmax);
        m_members |= HasPoint;
    }

    return m_point;
}

inline const foundation::Vector3d& ShadingPoint::get_offset_point(const foundation::Vector3d& direction) const
{
    assert(hit());
    assert(m_members & HasRefinedPoints);

    return foundation::dot(m_asm_geo_normal, direction) > 0.0 ? m_front_point : m_back_point;
}

inline const foundation::Vector3d& ShadingPoint::get_dpdu(const size_t uvset) const
{
    assert(hit());
    assert(uvset == 0);     // todo: support multiple UV sets

    if (!(m_members & HasPartialDerivatives))
    {
        compute_partial_derivatives();
        m_members |= HasPartialDerivatives;
    }

    return m_dpdu;
}

inline const foundation::Vector3d& ShadingPoint::get_dpdv(const size_t uvset) const
{
    assert(hit());
    assert(uvset == 0);     // todo: support multiple UV sets

    if (!(m_members & HasPartialDerivatives))
    {
        compute_partial_derivatives();
        m_members |= HasPartialDerivatives;
    }

    return m_dpdv;
}

inline const foundation::Vector3d& ShadingPoint::get_geometric_normal() const
{
    assert(hit());

    if (!(m_members & HasGeometricNormal))
    {
        if (m_primitive_type == PrimitiveTriangle)
        {
            if (m_members & HasWorldSpaceVertices)
            {
                // We already have the world space vertices of the hit triangle.
                // Use them to compute the geometric normal directly in world space.
                m_geometric_normal = foundation::cross(m_v1_w - m_v0_w, m_v2_w - m_v0_w);
            }
            else
            {
                cache_source_geometry();

                // Compute the object instance space geometric normal.
                const foundation::Vector3d v0(m_v0);
                const foundation::Vector3d v1(m_v1);
                const foundation::Vector3d v2(m_v2);
                m_geometric_normal = foundation::cross(v1 - v0, v2 - v0);

                // Transform the geometric normal to world space.
                m_geometric_normal =
                    m_assembly_instance_transform.normal_to_parent(
                        m_object_instance->get_transform().normal_to_parent(m_geometric_normal));
            }

            // Normalize the geometric normal.
            m_geometric_normal = foundation::normalize(m_geometric_normal);

            // Place the geometric normal in the same hemisphere as the original shading normal.
            if (foundation::dot(m_geometric_normal, get_original_shading_normal()) < 0.0)
                m_geometric_normal = -m_geometric_normal;

            // Remember which side of the geometric surface we hit.
            m_side =
                foundation::dot(m_ray.m_dir, m_geometric_normal) > 0.0
                    ? ObjectInstance::BackSide
                    : ObjectInstance::FrontSide;

            // Finally make the geometric normal face the direction of the incoming ray.
            if (m_side == ObjectInstance::BackSide)
                m_geometric_normal = -m_geometric_normal;
        }
        else
        {
            assert(m_primitive_type == PrimitiveCurve);

            // We assume flat ribbons facing incoming rays.
            m_geometric_normal = -normalize(m_ray.m_dir);
            m_side = ObjectInstance::FrontSide;
        }

        // The geometric normal is now available.
        m_members |= HasGeometricNormal;
    }

    return m_geometric_normal;
}

inline const foundation::Vector3d& ShadingPoint::get_shading_normal() const
{
    assert(hit());

    if (!(m_members & HasShadingNormal))
    {
        // Start with the original shading normal.
        m_shading_normal = get_original_shading_normal();

        if (m_primitive_type == PrimitiveTriangle)
        {
            // Apply the normal modifier if the material has one.
            const Material* material = get_material();
            if (material)
            {
                const INormalModifier* modifier = material->get_normal_modifier();
                if (modifier)
                {
                    m_shading_normal =
                        modifier->evaluate(
                            *m_texture_cache,
                            m_shading_normal,
                            get_uv(0),
                            get_dpdu(0),
                            get_dpdv(0));
                }
            }

            // Place the shading normal in the same hemisphere as the geometric normal.
            if (m_side == ObjectInstance::BackSide)
                m_shading_normal = -m_shading_normal;
        }

        // The shading normal is now available.
        m_members |= HasShadingNormal;
    }

    return m_shading_normal;
}

inline const foundation::Vector3d& ShadingPoint::get_original_shading_normal() const
{
    assert(hit());

    if (!(m_members & HasOriginalShadingNormal))
    {
        if (m_primitive_type == PrimitiveTriangle)
        {
            const double w = 1.0 - m_bary[0] - m_bary[1];

            if (m_members & HasWorldSpaceVertexNormals)
            {
                // We already have the world space vertex normals of the hit triangle.
                // Use them to compute the shading normal directly in world space.
                m_original_shading_normal =
                      m_n0_w * w
                    + m_n1_w * m_bary[0]
                    + m_n2_w * m_bary[1];
            }
            else
            {
                cache_source_geometry();

                // Compute the object instance space shading normal.
                m_original_shading_normal =
                      foundation::Vector3d(m_n0) * w
                    + foundation::Vector3d(m_n1) * m_bary[0]
                    + foundation::Vector3d(m_n2) * m_bary[1];

                // Transform the shading normal to world space.
                m_original_shading_normal =
                    m_assembly_instance_transform.normal_to_parent(
                        m_object_instance->get_transform().normal_to_parent(m_original_shading_normal));
            }

            // Normalize the shading normal.
            m_original_shading_normal = foundation::normalize(m_original_shading_normal);
        }
        else
        {
            assert(m_primitive_type == PrimitiveCurve);

            // We assume flat ribbons facing incoming rays.
            m_original_shading_normal = -normalize(m_ray.m_dir);
        }

        // The shading normal is now available.
        m_members |= HasOriginalShadingNormal;
    }

    return m_original_shading_normal;
}

inline const foundation::Basis3d& ShadingPoint::get_shading_basis() const
{
    assert(hit());

    if (!(m_members & HasShadingBasis))
    {
        // Construct the orthonormal basis.
        m_shading_basis.build(get_shading_normal(), normalize(get_dpdu(0)));

        // The orthonormal basis is now available.
        m_members |= HasShadingBasis;
    }

    return m_shading_basis;
}

inline ObjectInstance::Side ShadingPoint::get_side() const
{
    assert(hit());
    get_geometric_normal();
    return m_side;
}

inline const foundation::Vector3d& ShadingPoint::get_vertex(const size_t i) const
{
    assert(hit());
    assert(i < 3);

    if (!(m_members & HasWorldSpaceVertices))
    {
        cache_source_geometry();

        // Retrieve object instance space to assembly instance space transform.
        const foundation::Transformd& obj_instance_transform =
            m_object_instance->get_transform();

        // Transform triangle vertices to world space.
        const foundation::Vector3d v0(m_v0);
        const foundation::Vector3d v1(m_v1);
        const foundation::Vector3d v2(m_v2);
        m_v0_w = obj_instance_transform.point_to_parent(v0);
        m_v1_w = obj_instance_transform.point_to_parent(v1);
        m_v2_w = obj_instance_transform.point_to_parent(v2);
        m_v0_w = m_assembly_instance_transform.point_to_parent(m_v0_w);
        m_v1_w = m_assembly_instance_transform.point_to_parent(m_v1_w);
        m_v2_w = m_assembly_instance_transform.point_to_parent(m_v2_w);

        // World space triangle vertices are now available.
        m_members |= HasWorldSpaceVertices;
    }

    return (&m_v0_w)[i];
}

inline const foundation::Vector3d& ShadingPoint::get_vertex_normal(const size_t i) const
{
    assert(hit());
    assert(i < 3);

    if (!(m_members & HasWorldSpaceVertexNormals))
    {
        cache_source_geometry();

        // Retrieve instance space to assembly instance space transform.
        const foundation::Transformd& obj_instance_transform =
            m_object_instance->get_transform();

        // Transform vertex normals to world space.
        const foundation::Vector3d n0(m_n0);
        const foundation::Vector3d n1(m_n1);
        const foundation::Vector3d n2(m_n2);
        m_n0_w = obj_instance_transform.normal_to_parent(n0);
        m_n1_w = obj_instance_transform.normal_to_parent(n1);
        m_n2_w = obj_instance_transform.normal_to_parent(n2);
        m_n0_w = m_assembly_instance_transform.normal_to_parent(m_n0_w);
        m_n1_w = m_assembly_instance_transform.normal_to_parent(m_n1_w);
        m_n2_w = m_assembly_instance_transform.normal_to_parent(m_n2_w);

        // World space vertex normals are now available.
        m_members |= HasWorldSpaceVertexNormals;
    }

    return (&m_n0_w)[i];
}

inline const Material* ShadingPoint::get_material() const
{
    assert(hit());

    if (!(m_members & HasMaterial))
    {
        cache_source_geometry();

        m_material = 0;

        // Proceed with retrieving the material only if the hit triangle has one.
        if (m_primitive_pa != Triangle::None)
        {
            // Retrieve material indices from the object instance.
            const MaterialArray& materials =
                get_side() == ObjectInstance::BackSide
                    ? m_object_instance->get_back_materials()
                    : m_object_instance->get_front_materials();

            // Fetch the material.
            if (static_cast<size_t>(m_primitive_pa) < materials.size())
                m_material = materials[m_primitive_pa];
        }

        // The material at the intersection point is now available.
        m_members |= HasMaterial;
    }

    return m_material;
}

inline const AssemblyInstance& ShadingPoint::get_assembly_instance() const
{
    assert(hit());
    return *m_assembly_instance;
}

inline const Assembly& ShadingPoint::get_assembly() const
{
    assert(hit());
    cache_source_geometry();
    return *m_assembly;
}

inline const ObjectInstance& ShadingPoint::get_object_instance() const
{
    assert(hit());
    cache_source_geometry();
    return *m_object_instance;
}

inline const Object& ShadingPoint::get_object() const
{
    assert(hit());
    cache_source_geometry();
    return *m_object;
}

inline size_t ShadingPoint::get_object_instance_index() const
{
    assert(hit());
    return m_object_instance_index;
}

inline size_t ShadingPoint::get_region_index() const
{
    assert(hit());
    return m_region_index;
}

inline size_t ShadingPoint::get_primitive_index() const
{
    assert(hit());
    return m_primitive_index;
}

inline size_t ShadingPoint::get_primitive_attribute_index() const
{
    assert(hit());
    cache_source_geometry();
    return static_cast<size_t>(m_primitive_pa);
}

inline void ShadingPoint::cache_source_geometry() const
{
    if (!(m_members & HasSourceGeometry))
    {
        fetch_source_geometry();
        m_members |= HasSourceGeometry;
    }
}

inline void ShadingPoint::compute_partial_derivatives() const
{
    cache_source_geometry();

    const foundation::Vector3d& sn = get_original_shading_normal();

    if (m_primitive_type == PrimitiveTriangle)
    {
        const double du0 = static_cast<double>(m_v0_uv[0] - m_v2_uv[0]);
        const double dv0 = static_cast<double>(m_v0_uv[1] - m_v2_uv[1]);
        const double du1 = static_cast<double>(m_v1_uv[0] - m_v2_uv[0]);
        const double dv1 = static_cast<double>(m_v1_uv[1] - m_v2_uv[1]);

        const double det = du0 * dv1 - dv0 * du1;

        if (det == 0.0)
        {
            const foundation::Basis3d basis(sn);

            m_dpdu = basis.get_tangent_u();
            m_dpdv = basis.get_tangent_v();
        }
        else
        {
            const double rcp_det = 1.0 / det;

            const foundation::Vector3d dp0 = get_vertex(0) - get_vertex(2);
            const foundation::Vector3d dp1 = get_vertex(1) - get_vertex(2);

            m_dpdu = (dv1 * dp0 - dv0 * dp1) * rcp_det;
            m_dpdv = (du0 * dp1 - du1 * dp0) * rcp_det;

            m_dpdu = foundation::cross(sn, foundation::normalize(foundation::cross(m_dpdu, sn)));
            m_dpdv = foundation::cross(sn, foundation::normalize(foundation::cross(m_dpdv, sn)));
        }
    }
    else
    {
        assert(m_primitive_type == PrimitiveCurve);

        const CurveObject* curves = static_cast<const CurveObject*>(m_object);
        const CurveType& m_curve = curves->get_curve(m_primitive_index);

        const GScalar v = static_cast<GScalar>(m_bary[1]);
        const GVector3 tangent = m_curve.evaluate_tangent(v);

        m_dpdu = foundation::normalize(foundation::Vector3d(tangent));
        m_dpdv = foundation::normalize(foundation::cross(sn, m_dpdu));
    }
}

}       // namespace renderer

#endif  // !APPLESEED_RENDERER_KERNEL_SHADING_SHADINGPOINT_H
